import numpy as np
import pandas as pd
from scipy.interpolate import CubicSpline, PchipInterpolator, UnivariateSpline
from typing import Optional, Sequence



class DataManipulation:
    VERSION = 'Kantris.DataManipulation: 0.1.2'
    @staticmethod
    def LinearInterpol(x, xp, fp):
        """"
        x: target x-values
        xp: x-values of data
        fp: y-values of data
        """
        return np.interp(x, xp, fp)
    @staticmethod
    def CubicInterpol(x, xp, fp, bc_type='natural', extrapolate=True):
        if not np.all(np.diff(xp) > 0):
            # Sortiere xp und fp gemeinsam nach xp
            idx_sort = np.argsort(xp)
            xp = xp[idx_sort]
            fp = fp[idx_sort]


        cs = CubicSpline(xp, fp, bc_type=bc_type, extrapolate=extrapolate)
        return cs(x)
    @staticmethod
    def HermiteInterpol(x, xp, fp, extrapolate=True):
        pchip = PchipInterpolator(xp, fp, extrapolate=extrapolate)
        return pchip(x)
    @staticmethod
    def UnivariateInterpol(x, xp, fp, s):
        pchip = UnivariateSpline(xp, fp, s=s)
        return pchip(x)
    

class Array:
    """
    Methods:
        Col
    """
    VERSION = 'Kantris.Array: 0.1.1'


    @staticmethod
    def Merge(*arrays: np.ndarray) -> np.ndarray:
        """
        Nimmt beliebig viele NumPy-Arrays (1D oder 2D) und gibt ein 2D-Array zurück,
        in dem alle 1D-Eingaben als Spalten angeordnet sind. Falls ein Input 2D ist,
        werden seine Spalten jeweils als separate 1D-Arrays behandelt.

        Parameters
        ----------
        *arrays : np.ndarray
            Eine beliebige Anzahl von 1D- oder 2D-NumPy-Arrays.

        Returns
        -------
        np.ndarray
            Ein 2D-Array der Form (L, N), wobei L die gemeinsame Länge aller
            Spalten und N die Gesamtanzahl der resultierenden Spalten ist.

        Raises
        ------
        ValueError
            Wenn kein Array übergeben wird, ein Array weniger als 1D bzw. mehr als 2D ist
            oder die Längen der resultierenden Spalten nicht übereinstimmen.
        """
        if not arrays:
            raise ValueError("Mindestens ein Array muss übergeben werden.")

        cols = []
        for arr in arrays:
            arr = np.asarray(arr)
            if arr.ndim == 1:
                cols.append(arr)
            elif arr.ndim == 2:
                # Spalten einzeln hinzufügen
                for j in range(arr.shape[1]):
                    cols.append(arr[:, j])
            else:
                raise ValueError(
                    f"Eingabe mit ndim={arr.ndim} wird nicht unterstützt (nur 1D oder 2D).")
        
        # Prüfen, ob wirklich 1D-Spalten vorliegen und gleiche Länge haben
        lengths = [col.shape[0] for col in cols]
        if len(set(lengths)) != 1:
            raise ValueError(
                "Alle resultierenden Spalten müssen dieselbe Länge haben.")

        # Spaltenweises Zusammenfügen
        return np.column_stack(cols)

    
    @staticmethod
    def Col(array: np.ndarray, col: int) -> np.ndarray:
        return array[:, col]

    @staticmethod
    def Normalise(array, mode: str = "standard", col: int|tuple[int]|list[int] = 0) -> np.ndarray:
        """
        Normalisiert bestimmte Spalten eines Arrays nach verschiedenen Methoden.

        Args:
            array (np.ndarray): Eingabedaten.
            columns (list[str] | str): Liste der zu normalisierenden Spaltenindexe.
            Arg (str): Optionen:
                - "standard": (default) Mittelwert = 0, Std-Abw = 1
                - "mean": nur Mittelwertzentrierung
                - "start": Startwert auf 0 setzen
                - "amplitude": Skaliere Werte auf [0, 1]

        Returns:
            np.ndarray: Neues Array mit normalisierten Spalten.
        """
        if isinstance(col, (tuple, list)):
            for index in col:
                if not isinstance(index, int):
                    print(index, 'is invalid column index')
                    return array
                array = Array.Normalise(array, mode, index)
            return array
        
        elif isinstance(col, int):
            column = Array.Col(array, col)
            if mode in ('standard', 'standardise', 'std'):
                mean = column.mean()
                std = column.std()
                if std != 0:
                    column = (column - mean) / std

            elif mode in ('mean',):
                column = column - column.mean()
            
            elif mode in ('start',):
                column = column - column[0]
            
            elif mode in ('normal', 'normalise', 'norm'):
                min = column.min()
                max = column.max()
                column = column - min
                column = column/(max-min)

            else:
                print(mode, 'is invalid normalisation mode')
                return array
            
            array[:, col] = column
            return array
        
        else:
            print(col, 'is invalid column index')
            return array

    @staticmethod
    def Scale(array, scale: float, col: int|tuple[int]|list[int] = 0) -> np.ndarray:
        if isinstance(col, (tuple, list)):
            for index in col:
                if not isinstance(index, int):
                    print(index, 'is invalid column index')
                    return array
                array = Array.Scale(array, index)
            return array
        
        elif isinstance(col, int):
            array[:, col] = array[:, col]*scale
            return array
        else:
            print(col, 'is invalid column index')
            return array
    def RemoveOutliers(
        data: np.ndarray,
        threshold: float,
        cols: Optional[Sequence[int]] = None
    ) -> np.ndarray:
        """
        Entfernt alle Zeilen aus `data`, in denen in den angegebenen Spalten
        mindestens ein Wert mehr als `threshold` Standardabweichungen vom 
        jeweiligen Spaltenmittelwert abweicht.

        Parameters
        ----------
        data : np.ndarray
            2D-Array mit numerischen Werten, Form (n_samples, n_features).
        threshold : float
            Anzahl der Standardabweichungen, ab der ein Wert als Ausreißer gilt.
        cols : Sequence[int], optional
            Liste der Spalten-Indizes, die auf Ausreißer geprüft werden sollen.
            Standardmäßig (None) wird jede Spalte geprüft.

        Returns
        -------
        np.ndarray
            Gefiltertes 2D-Array mit allen "ausreißerfreien" Zeilen.
        """
        n_samples, n_features = data.shape

        # Falls keine Spalten angegeben, prüfen wir alle
        if cols is None:
            cols = np.arange(n_features)
        else:
            # in np-Array umwandeln, damit data[:, cols] funktioniert
            cols = np.array(cols, dtype=int)

        # Nur die zu prüfenden Spalten extrahieren
        sub = data[:, cols]

        # Mittelwert und Standardabweichung der Submatrix berechnen
        mean = np.mean(sub, axis=0)
        std  = np.std(sub,  axis=0)

        # Absolute Abweichung jeder Zeile in den gewählten Spalten vom Mittelwert
        deviation = np.abs(sub - mean)

        # Maske: True, wenn in ALLEN geprüften Spalten Abweichung ≤ threshold·std
        keep_mask = np.all(deviation <= threshold * std, axis=1)

        # Gefiltertes Array zurückgeben (inkl. unveränderter Spalten)
        return data[keep_mask]


class Dataframe:
    """
    Methods:
        ReadFromCSV
        ColumnToArray
        AddCol
        Normalise
    """
    VERSION = 'Kantris.Dataframe: 0.1.0'


    @staticmethod
    def Functions():
        text = """
        ReadFromCSV()
        ColumnToArray()
        AddCol()
        Normalise()
        """
        print(text)

    @staticmethod
    def ReadFromCSV(filepath: str, sep: str = ',') -> pd.DataFrame:
        """
        Liest eine CSV-Datei ein (erste Zeile als Header) und gibt ein pandas DataFrame zurück.

        Args:
            filepath (str): Pfad zur CSV-Datei.
            sep (str, default = ','): Seperator der Datei.

        Returns:
            pd.DataFrame: DataFrame mit den eingelesenen Daten.
        """
        return pd.read_csv(filepath, sep=sep)
    
    @staticmethod
    def ColumnToArray(df: pd.DataFrame, columns: list[str]) -> np.ndarray:
        """
        Liest die angegebenen Spalten aus dem DataFrame aus und gibt sie als NumPy-Array zurück.

        Args:
            df (pd.DataFrame): Das DataFrame, aus dem die Spalten entnommen werden sollen.
            columns (list[str]): Liste der Spaltennamen, die extrahiert werden sollen.

        Returns:
            np.ndarray: 2D-Array mit den Werten der ausgewählten Spalten (Form: (n_rows, len(columns))).
        """
        return df[columns].to_numpy()
    
    @staticmethod
    def AddCol(
        df: pd.DataFrame,
        column: np.ndarray,
        name: str = None,
        interpol: tuple = None
    ) -> pd.DataFrame:
        """
        Fügt dem DataFrame df eine neue Spalte hinzu mit ggf. Interpolation auf eine Spalte.

        Args:
            df (pd.DataFrame): Das Ziel-DataFrame, dem eine Spalte hinzugefügt wird.
            column (np.ndarray): 1D-Array mit den Werten, die als neue Spalte eingefügt werden sollen.
            name (str, optional): Name der neuen Spalte. 
                                  Wenn None, wird der nächste Index (als String) verwendet.
            interpol (tuple, optional): Tuple der Form
                (spaltenname: str, x_vals: np.ndarray, [method: str]).
                - spaltenname: Name der existierenden Spalte in df, deren Werte (y) als Stützwerte dienen.
                - x_vals: 1D-Array von x-Koordinaten, die zu df[spaltenname] gehören.
                - method (optional): 'linear' oder 'cubic' (Standard: 'linear').

                Wenn interpol angegeben ist, wird zunächst eine Interpolation 
                von df[spaltenname] an den Stützstellen x_vals auf die neuen x-Werte 
                column vorgenommen. Das Ergebnis dieser Interpolation wird 
                dann als neue Spalte eingesetzt.

        Returns:
            pd.DataFrame: Das DataFrame mit der hinzugefügten Spalte (inplace).
        """
        # 1. Überprüfen, dass column 1D ist
        column = np.asarray(column)
        if column.ndim != 1:
            raise ValueError("Array for Column must be of Dimension 1")

        # 2. Interpolations-Logik
        if interpol is not None:
            if not (isinstance(interpol, tuple) and 2 <= len(interpol) <= 3):
                raise ValueError(
                    "interpol muss ein Tuple sein: "
                    "(spaltenname: str, x_vals: np.ndarray, [method: str])"
                )
            spaltenname = interpol[0]
            x_vals = np.asarray(interpol[1])
            method = 'linear' if len(interpol) == 2 else interpol[2]

            # Spaltenname existiert?
            if spaltenname not in df.columns:
                raise KeyError(f"Spalte '{spaltenname}' nicht im DataFrame gefunden.")

            y_vals = df[spaltenname].to_numpy()
            if y_vals.ndim != 1:
                raise ValueError(f"Spalte '{spaltenname}' muss 1D sein.")

            # Überprüfen, dass x_vals 1D und gleich lang wie y_vals ist
            if x_vals.ndim != 1:
                raise ValueError(
                    "x_vals muss 1D sein."
                )

            # Interpolation durchführen
            if method == 'linear':
                new_vals = DataManipulation.LinearInterpol(y_vals, x_vals, column)
            elif method == 'cubic':
                new_vals = DataManipulation.CubicInterpol(y_vals, x_vals, column)
            else:
                raise ValueError("Unbekannte Methode in interpol: nutze 'linear' oder 'cubic'.")

            # Das interpolierte Ergebnis ist unsere neue Spalte
            new_column = new_vals
        else:
            # Keine Interpolation: direkte Verwendung des Arrays
            new_column = column

        # 3. Spaltenname festlegen
        if name is None:
            # Nächster Index als String
            next_index = len(df.columns)
            name = str(next_index)

        # 4. Länge prüfen: new_column muss so viele Einträge haben wie df Zeilen
        if new_column.shape[0] != df.shape[0]:
            raise ValueError(
                f"Neue Spalte hat Länge {new_column.shape[0]}, "
                f"muss aber {df.shape[0]} entsprechen."
            )

        # 5. Spalte hinzufügen (inplace)
        df[name] = new_column
        return df
    
    @staticmethod
    def Normalise(df: pd.DataFrame, columns, Arg: str = 'standard') -> pd.DataFrame:
        """
        Normalisiert bestimmte Spalten eines DataFrames nach verschiedenen Methoden.

        Args:
            df (pd.DataFrame): Eingabedaten.
            columns (list[str] | str): Liste der zu normalisierenden Spaltennamen oder '*' für alle Spalten (falls '*' kein Spaltenname ist).
            Arg (str): Optionen:
                - "standard": (default) Mittelwert = 0, Std-Abw = 1
                - "mean": nur Mittelwertzentrierung
                - "start": Startwert auf 0 setzen
                - "amplitude": Skaliere Werte auf [0, 1]

        Returns:
            pd.DataFrame: Neues DataFrame mit normalisierten Spalten.
        """

        df = df.copy()

        # Spaltenauswahl
        if columns == '*' and '*' not in df.columns:
            selected_columns = df.columns
        elif isinstance(columns, str):
            selected_columns = [columns]
        else:
            selected_columns = columns

        for col in selected_columns:
            x = df[col].values.astype(float)

            if Arg == "amplitude":
                min_val = np.min(x)
                max_val = np.max(x)
                if max_val - min_val != 0:
                    x = (x - min_val) / (max_val - min_val)
                else:
                    x[:] = 0.0  # falls konstante Spalte

            elif Arg == "start":
                x = x - x[0]

            elif Arg == "mean":
                x = x - np.mean(x)

            elif Arg == "standard":
                mean = np.mean(x)
                std = np.std(x)
                if std != 0:
                    x = (x - mean) / std
                else:
                    x[:] = 0.0  # falls konstante Spalte

            df[col] = x

        return df


import numpy as np
from scipy.integrate import odeint
from typing import Sequence, Callable, Optional, Any

class Calc:
    VERSION = 'Kantris.Calc: 0.1.3'
    class ODE:
        """
        Numerically solves a System of Ordinary Differential Equations\n
        T() Sets a List of the values of T to be calculated \n
        Func() Sets the Function to be solved for. Must be in format func(y, t, *Args) \n
        Args() Sets the arguments for Func() \n
        Y0() Sets the starting paramters. Must match the Dimension of Func() \n
        Run() solves the ODE and returns a list of lists in format [t, y1, y2, ...] \n
        """
        def __init__(self):
            self.t: Optional[Sequence[float]] = None
            self.y0: Optional[Sequence[float]] = None
            self.args: tuple[Any, ...] = ()
            self.func: Optional[Callable[..., Sequence[float]]] = None
    
        def T(self, t: Sequence[float]):
            self.t = t
            return self
        def Y0(self, y0: Sequence[float]):
            self.y0 = y0
            return self
        def Args(self, args: tuple[Any, ...]):
            self.args = args
            return self
        def Func(self, func: Callable[..., Sequence[float]]):
            """
            Example for a 2-Dimensional ODE with 2 Args/Variables (harmonic oszilator with friction): \n
            def diffEq(y, t, A, B):
                d1, d2 = y
                dydt = [d2, -A*d1 - B*d2]
                return dydt
            """
            self.func = func
            return self


        def Run(self) -> np.ndarray:
            # Check required fields
            missing = [f for f in ("t", "y0", "func") if getattr(self, f) is None]
            if missing:
                raise RuntimeError(f"Missing parameters: {', '.join(missing)}")
            # Solve
            sol = odeint(self.func, self.y0, self.t, args=self.args)
            # Optional: return t alongside solution
            return np.column_stack((self.t, sol))


import numpy as np
import matplotlib
matplotlib.use("Agg")  # wichtig: muss VOR pyplot import stehen

import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, AutoMinorLocator, FixedLocator
from matplotlib.patches import FancyArrowPatch

from .Basics import Basics
from .DataManipulation import Array

import pandas as pd
import scipy
import scipy.optimize


class Plotter:
    VERSION = "0.3.0"
    class Legend:
        def __init__(self, show = True):
            self.show = show
            self.position = "best"
        
        def Position(self, value):
            self.position = value
            return self

    class Graph:
        def __init__(self, x_values: list | np.ndarray | tuple | None = None, y_values: list | np.ndarray | tuple | None = None):
            if x_values is  None:
                self.x_values = []
            else:
                self.x_values = x_values
            if y_values is  None:
                self.y_values = []
            else:
                self.y_values = y_values
            self.x_range = None
            self.y_range = None

            self.x_mirror = False
            self.y_mirror = False
            self.xy_mirror = False
            self.z = None
            self.segmented = False
            self.segmented_jump = 0

        def Points(self, points: np.ndarray):
            if points.ndim != 2:
                print(f"Dimension is not 2 but {points.dim}")
                return self
            self.x_values = points[:, 0]
            self.y_values = points[:, 1]
            return self
        def X(self, x_values: list|np.ndarray|tuple):
            self.x_values = x_values
            return self
        def Y(self, y_values: list|np.ndarray|tuple):
            self.y_values = y_values
            return self
        def Z(self, value):
            self.z = value
            return self
        def Segmented(self, value):
            if value is False:
                self.segmented = False
            else:
                self.segmented = True
                if value <= 0:
                    value = 0.1
                    print('INFO for Plotter.Segmented: Jump cannot be <= 0. Set to 0.1')
                self.segmented_jump = value
            return self
        def xRange(self, min, max):
            self.x_range = (min, max)
            return self
        def yRange(self, min, max):
            self.y_range = (min, max)
            return self
        def Mirror(self, which="x"):
            if which == "x":
                self.x_mirror = True
            elif which == "y":
                self.y_mirror = True
            elif which == "xy":
                self.xy_mirror = True
            return self

    class Scatter:
        def __init__(self, x_values: list | np.ndarray | tuple | None = None, y_values: list | np.ndarray | tuple | None = None):
            if x_values is  None:
                self.x_values = []
            else:
                self.x_values = x_values
            if y_values is  None:
                self.y_values = []
            else:
                self.y_values = y_values
            self.x_range = None
            self.y_range = None

            self.x_range = None
            self.y_range = None

            self.x_mirror = False
            self.y_mirror = False
            self.xy_mirror = False
            self.z = None

        def Points(self, points: np.ndarray):
            if points.ndim != 2:
                print(f"Dimension is not 2 but {points.dim}")
                return self
            self.x_values = points[:, 0]
            self.y_values = points[:, 1]
            return self
        def Z(self, value):
            self.z = value
            return self
        def xRange(self, min, max):
            self.x_range = (min, max)
            return self
        def yRange(self, min, max):
            self.y_range = (min, max)
            return self
        def Mirror(self, which="x"):
            if which == "x":
                self.x_mirror = True
            elif which == "y":
                self.y_mirror = True
            elif which == "xy":
                self.xy_mirror = True
            return self

    class Line:
        """
        Defines Marker-Styling for used in Graph-Style Plots.\\
        Attributes are:\\
        Color()\\
        Width()\\
        Style()\\
        """
        def __init__(self):
            self.color = None
            self.width = 2
            self.style = "-"

        def Color(self, value):
            self.color = value
            return self
        def Width(self, value):
            self.width = value
            return self
        def Style(self, value):
            """
            Sets the LineStyle.
            Values are:
            "-", "--"
            """
            self.style = value
            return self
        
    class Marker:
        """
        Defines Marker-Styling for used in Graph- or Scatter-Style Plots.\\
        Attributes are:\\
        Color()\\
        Alpha()\\
        Size()\\
        Style()\\
        Step()
        """
        def __init__(self):
            self.color = None
            self.size = 10
            self.style = "o"
            self.step = None
            self.alpha = 1

        def Color(self, value):
            self.color = value
            return self
        def Alpha(self, value):
            self.alpha = value
            return self
        def Size(self, value):
            self.size = value
            return self
        def Style(self, value):
            self.style = value
            return self
        def Step(self, value):
            self.step = value
            return self
        
    class Label:
        """
        Defines Label-Styling and Text. Universal.\\
        Attributes are:\\
        Text()\\
        Unit()\\
        Color()\\
        Fontsize()\\
        Enumerate()\\
        String() - Returns the String Representation
        """
        def __init__(self, text: str = '', latex: bool = False, enumerate: bool = None):
            self.text = text
            self.text_latex = latex
            self.color = "black"
            self.fontsize = 16
            self.unit = ""
            self.unit_latex = True
            self.unit_brackets = True
            self.enumerate = enumerate

        def Text(self, value, latex = False):
            self.text = value
            self.text_latex = latex
            return self

        def Unit(self, value, latex = True, brackets = True):
            self.unit = value
            self.unit_latex = latex
            self.unit_brackets = brackets
            return self

        def Color(self, value):
            self.color = value
            return self
        
        def Fontsize(self, value):
            self.fontsize = value
            return self

        def Enumerate(self, callsign, value):
            self.enumerate = (callsign, value)
            return self

        def String(self, enumerate: bool = False):
            string = ""
            if self.text_latex:
                string = f'$\\mathrm{{{self.text}}}$'
            else:
                string = self.text
            if self.unit != "" and self.unit is not None:
                if self.unit_latex and self.unit_brackets:
                    string = f'{string} [$\\mathrm{{{self.unit}}}$]'
                elif self.unit_latex and not self.unit_brackets:
                    string = f'{string} $\\mathrm{{{self.unit}}}$'
                elif self.unit_brackets and not self.unit_latex:
                    string = f'{string} [{self.unit}]'
                else:
                    string = f'{string} {self.unit}'
            if string == "" and enumerate and self.enumerate:
                return f'{self.enumerate[0]} {self.enumerate[1]}'
            return string
        
    class TickTicker:
        def __init__(self):
            self = self.DefaultMajor()


        def DefaultMajor(self):
            self.length = 8
            self.width = 1
            self.color = "black"
            self.direction = "inout"
            return self
        
        def DefaultMinor(self):
            self.length = 4
            self.width = 1
            self.color = "black"
            self.direction = "inout"
            return self
        
        def Length(self, value):
            self.length = value
            return self
        def Width(self, value):
            self.width = value
            return self
        def Color(self, value):
            self.color = value
            return self
        def Direction(self, value):
            self.direction = value
            return self
        
    class TickGrid:
        def __init__(self):
            self = self.DefaultMajor()
        
        def DefaultMajor(self):
            self.color = "gray"
            self.alpha = 0.25
            self.linewidth = 1
            self.linestyle = "-"
            return self
        
        def DefaultMinor(self):
            self.color = "gray"
            self.alpha = 0.2
            self.linewidth = 0.6
            self.linestyle = "--"
            return self

        def Color(self, value):
            self.color = value
            return self
        def Alpha(self, value):
            self.alpha = value
            return self
        def Width(self, value):
            self.linewidth = value
            return self
        def Style(self, value):
            self.linestyle = value
            return self

    class ColorCycler:
        def __init__(self):
            self.colors = [f'C{i}' for i in range(0,10)]
            self.counter = 0
        def increase(self):
            if self.counter < 9:
                self.counter = self.counter +1
            else:
                self.counter = 0
        def GetCurrent(self):
            return self.colors[self.counter]
        def GetCurrentAndCycle(self):
            color = self.colors[self.counter]
            self.increase()
            return color
        
    class Objects:
        class Vector:
            """
            Defines a Vector to be used in a Plots 'OBJECTS'\\
            Attributes are:\\
            Vector() - Sets the Vector as an 2-Dim np.ndarray. Length will be scaled to length of input vector\\
            Rad() - Sets the Vector for the input angle in radians. Keeps current length\\
            Deg() - Sets the Vector for the input angle in degree. Keeps current length\\
            Length() - Sets length. Unit may be arbitrary\\
            Origin() - Sets the origin coordinates.\\
            Color()\\
            Width()\\
            Headwidth()\\
            Headlenght()\\
            """
            def __init__(self, vector: np.ndarray = np.ndarray([0,0])):
                self.length = np.linalg.norm(vector)
                if self.length != 0:
                    self.vector = vector/self.length
                else:
                    self.vector = vector
                self.origin = (0,0)

                self.width = 1
                self.headwidth = 2
                self.headlength = 2

                self.color = None
                self.alpha = 1

                self.mirror = False


            def Vector(self, value):
                self.vector = value
                self.length = np.linalg.norm(self.vector)
                return self
            def Rad(self, angle, length=None):
                if length:
                    self.length = length
                self.vector = np.array([np.cos(angle), np.sin(angle)])
                return self
            def Deg(self, angle, length=None):
                if length:
                    self.length = length
                angle = angle/360*2*np.pi
                self.vector = 5*np.array([np.cos(angle), np.sin(angle)])
                return self
            def Length(self, value):
                self.length = value
                return self
            def Origin(self, x, y):
                self.origin = (x, y)
                return self
            def Color(self, value):
                self.color = value
                return self
            def Alpha(self, value):
                self.alpha = value
                return self
            def Width(self, value):
                self.width = value
                return self
            def HeadWidth(self, value):
                self.headwidth = value
                return self
            def HeadLength(self, value):
                self.headlength = value
                return self
            def Mirror(self, value=True):
                self.mirror = value
                return self

    @staticmethod
    def Plot(config: dict, graphs:list[dict], objects: list[dict] | None = None) -> plt.Figure|None:
        """
        Creates a Graph-Plot based on the Config and Graphs given:
        - data: DataFrame or List of Lists
        - config: Config Dictionary
        - graphs: List of Dictionary for Graphs to be plotted
        - objects: Optional, List of Dictionary for Objects (Labels, Vectors etc.) to be plotted
        
        Example Plot-Config can be found here: https://1.1.1.1 or printed with Plotter.Config("plot"), Plotter.MinimalConfig("plot") or Plotter.FullConfig("plot")\\
        Example Graph-Config can be found here: https://1.1.1.1 or printed with Plotter.Config("graph"), Plotter.MinimalConfig("graph") or Plotter.FullConfig("graph")\\
        Example Object-Config can be found here: https://1.1.1.1 or printed with Plotter.Config("object"), Plotter.MinimalConfig("object") or Plotter.FullConfig("object")
        """

        if objects is None:
            objects = []

        

        plot_conf = config.get("plot", {})
        plot_size = Basics.dictMultiGet(plot_conf, ('figsize', 'plotsize', 'size'), (12,6))
        if plot_size == '1/1':
            plot_size = (10, 10)
        elif plot_size == '1/2':
            plot_size == (20,10)
        elif plot_size == '1/3':
            plot_size == (30,10)
        elif plot_size == '1/4':
            plot_size == (40,10)
        filename = Basics.dictMultiGet(plot_conf, ('filename', 'file', 'save', 'saveas'), None)
        bg_color = Basics.dictMultiGet(plot_conf, ('color', 'colour'), None)
        border_color = Basics.dictMultiGet(plot_conf, ('edge-color', 'edge-colour'), None)

        x_axis_conf = config.get("x-axis", {})
        y_axis_conf = config.get("y-axis", {})

        Log = BoolLogger()
        Log.Set(Basics.dictMultiGet(plot_conf, ('logging', 'log'), False))
        fig, ax = plt.subplots(figsize=plot_size)

        if bg_color:
            fig.patch.set_facecolor(bg_color)
            ax.set_facecolor(bg_color)
        if border_color:
            for spine in ax.spines.values():
                spine.set_edgecolor(border_color)

        # Achsenbeschriftungen inkl. Einheiten (Latex-Formatierung möglich)
        title = Basics.dictMultiGet(plot_conf, ('title', 'plottitle', 'plot title', 'name', 'plotname', 'plot name'), 'Plot')
        if isinstance(title, str):
            title = Plotter.Label(title)
        elif isinstance(title, Plotter.Label):
            pass
        else:
            print('error on Plotter.Plot: plot/title is not string or Plotter.Label')
            return
        ax.set_title(title.String(), fontsize=title.fontsize, color=title.color)

        xlabel = Basics.dictMultiGet(x_axis_conf, ('label', 'name'), 'x-Axis')
        if isinstance(xlabel, str):
            xlabel = Plotter.Label(xlabel)
        elif isinstance(xlabel, Plotter.Label):
            pass
        else:
            print('error on Plotter.Plot: plot/x-axis/label is not string or Plotter.Label')
            return
        ax.set_xlabel(xlabel.String(), fontsize=xlabel.fontsize, color=xlabel.color)

        ylabel = Basics.dictMultiGet(y_axis_conf, ('label', 'name'), 'y-Axis')
        if isinstance(ylabel, str):
            ylabel = Plotter.Label(ylabel)
        elif isinstance(ylabel, Plotter.Label):
            pass
        else:
            print('error on Plotter.Plot: plot/y-axis/label is not string or Plotter.Label')
            return
        ax.set_ylabel(ylabel.String(), fontsize=ylabel.fontsize, color=ylabel.color)

        ColorCyler = Plotter.ColorCycler()

        for i, graph_config in enumerate(graphs):
            graph_label = Basics.dictMultiGet(graph_config, ('label', 'name'), '')
            if isinstance(graph_label, str):
                graph_label = Plotter.Label(graph_label)
            elif isinstance(graph_label, Plotter.Label):
                pass
            else:
                print(f'error on Plotter.Plot: graphs {i}/label is not string or Plotter.Label')
                return
            graph_label.Enumerate('Graph', i)
            graph_data = graph_config.get("data", None)
            if graph_data is None:
                print(f"Skipping {graph_label}, Data is None")

            if isinstance(graph_data, Plotter.Graph):
                #Values for Graph Instance
                mask= CreateMask(graph_data.x_values, graph_data.x_range)
                x_values_plot = graph_data.x_values[mask]
                y_values_plot = graph_data.y_values[mask]
                mask= CreateMask(y_values_plot, graph_data.y_range)
                x_values_plot = x_values_plot[mask]
                y_values_plot = y_values_plot[mask]
                
                #Style for Graph Instance
                graph_linestyling = graph_config.get("line-styling", Plotter.Line())
                graph_markerstyling = graph_config.get("marker-styling", Plotter.Marker().Style(""))
                if graph_linestyling.color is None:
                    graph_linestyling.color = ColorCyler.GetCurrentAndCycle()
                    if graph_markerstyling.color is None:
                        graph_markerstyling.color = graph_linestyling.color
                elif graph_markerstyling.color is None:
                    graph_markerstyling.color = graph_linestyling.color

                #Plot
                if not graph_data.segmented:
                    ax.plot(x_values_plot, y_values_plot, label=graph_label.String(enumerate=True), color=graph_linestyling.color, linewidth = graph_linestyling.width, linestyle=graph_linestyling.style ,marker=graph_markerstyling.style, markerfacecolor=graph_markerstyling.color, markersize=graph_markerstyling.size, zorder=graph_data.z)
                    if graph_data.x_mirror:
                        ax.plot(-x_values_plot, y_values_plot, label=None, color=graph_linestyling.color, linewidth = graph_linestyling.width, linestyle=graph_linestyling.style ,marker=graph_markerstyling.style, markerfacecolor=graph_markerstyling.color, markersize=graph_markerstyling.size, zorder=graph_data.z)
                    if graph_data.y_mirror:
                        ax.plot(x_values_plot, -y_values_plot, label=None, color=graph_linestyling.color, linewidth = graph_linestyling.width, linestyle=graph_linestyling.style ,marker=graph_markerstyling.style, markerfacecolor=graph_markerstyling.color, markersize=graph_markerstyling.size, zorder=graph_data.z)
                else:
                    graph_points = Array.Merge(x_values_plot, y_values_plot)
                    graph_segments = Basics.SplitSegments(graph_points, graph_data.segmented_jump)
                    for i, seg in enumerate(graph_segments):
                        if i == 0:
                            ax.plot(Array.Col(seg, 0), Array.Col(seg, 1), label=graph_label.String(enumerate=True), color=graph_linestyling.color, linewidth = graph_linestyling.width, linestyle=graph_linestyling.style ,marker=graph_markerstyling.style, markerfacecolor=graph_markerstyling.color, markersize=graph_markerstyling.size, zorder=graph_data.z)
                        else:
                            ax.plot(Array.Col(seg, 0), Array.Col(seg, 1), label=None, color=graph_linestyling.color, linewidth = graph_linestyling.width, linestyle=graph_linestyling.style ,marker=graph_markerstyling.style, markerfacecolor=graph_markerstyling.color, markersize=graph_markerstyling.size, zorder=graph_data.z)
                #Fits
                graph_fits = Basics.dictMultiGet(graph_config, ('fit', 'fits'), [])
                if isinstance(graph_fits, dict):
                    graph_fits = [graph_fits]

                for j, fit in enumerate(graph_fits):
                    fit_label = Basics.dictMultiGet(fit, ('label', 'name'), '')
                    if isinstance(fit_label, str):
                        fit_label = Plotter.Label(fit_label)
                    elif isinstance(fit_label, Plotter.Label):
                        pass
                    else:
                        print(f'error on Plotter.Plot: graphs {i}/fit {j}/label  is not string or Plotter.Label')
                        return
                    
                    fit_bounds = ([],[])
                    fit_bounds_c = Basics.dictMultiGet(fit, ('bounds', 'fit-bounds'), {})
                    fit_fixed_params = Basics.dictMultiGet(fit, ('fixed', 'fit-fixed'), None)
                    fit_label.Enumerate('Fit', j)

                    fit_type = fit.get("type", None)
                    fit_range = fit.get("fit-range", None)
                    fit_plot_range = fit.get("plot-range", None)
                    fit_function = None
                    popt = None
                    popt_map = {}
                    perr = None
                    perr_map = {}
                    if fit_type is None:
                        continue
                    mask = CreateMask(graph_data.x_values, fit_range)
                    x_values_fit = graph_data.x_values[mask]
                    y_values_fit = graph_data.y_values[mask]

                    if fit_type.lower() in ["lin", "linear"]:
                        fitKeys = ['m', 'B']
                        fit_bounds = {
                            'm': (-np.inf, np.inf),
                            'B': (-np.inf, np.inf),
                        }
                        for bound_key, bounds_item in fit_bounds_c.items():
                            fit_bounds[bound_key] = bounds_item
                            
                        
                        A0 = (y_values_fit[-1] - y_values_fit[0])/(x_values_fit[-1] - x_values_fit[0])
                        B0 = np.min(y_values_fit)
                        def fitfunc(x, A, B):
                            return A*x + B
                        popt, pcov = scipy.optimize.curve_fit(fitfunc, x_values_fit, y_values_fit, bounds=fit_bounds)
                        perr = np.sqrt(np.diag(pcov))

                        for popt_i, popt_k in enumerate(fitKeys):
                            popt_map[popt_k] = Basics.roundFloat(popt[popt_i])
                        for perr_i, perr_k in enumerate(fitKeys):
                            perr_map[perr_k] = Basics.roundFloat(perr[perr_i])
                        fit_function = fitfunc
                        
                    elif fit_type.lower() in ["gauss", "gaus"]:
                        fitKeys = ['A', 'mu', 'sigma']
                        fit_bounds = ([-np.inf, -np.inf, -np.inf], [np.inf, np.inf, np.inf])
                        for bound_key, bounds_item in fit_bounds_c.items():
                            l = fitKeys.index(bound_key)
                            fit_bounds[0][i] = bounds_item[0]
                            fit_bounds[1][i] = bounds_item[1]

                        A0 = np.max(y_values_fit)
                        mu0 = (x_values_fit[-1] + x_values_fit[0])/2
                        sigma0 = (x_values_fit[-1] - x_values_fit[0])/3
                        p0 = [A0, mu0, sigma0]
                        def fitfunc(x, A, mu, sigma):
                            return A*np.exp(-1/2*((x-mu)**2)/(sigma**2))
                        popt, pcov = scipy.optimize.curve_fit(fitfunc, x_values_fit, y_values_fit, p0=p0, bounds=fit_bounds)
                        perr = np.sqrt(np.diag(pcov))
                        for popt_i, popt_k in enumerate(fitKeys):
                            popt_map[popt_k] = Basics.roundFloat(popt[popt_i])
                        for perr_i, perr_k in enumerate(fitKeys):
                            perr_map[perr_k] = Basics.roundFloat(perr[perr_i])
                        fit_function = fitfunc
                        area = (popt_map['A'])*np.sqrt(2*np.pi)*popt_map['sigma']
                        popt_map['area'] =  Basics.roundFloat(area)
                        fitKeys.append('area')

                    elif fit_type.lower() in ["gausslin", "gauslin"]:
                        fitKeys = ['A', 'mu', 'sigma', 'B', 'C']
                        fit_bounds = ([-np.inf, -np.inf, -np.inf, -np.inf, -np.inf], [np.inf, np.inf, np.inf, np.inf, np.inf])
                        for bound_key, bounds_item in fit_bounds_c.items():
                            l = fitKeys.index(bound_key)
                            fit_bounds[0][l] = bounds_item[0]
                            fit_bounds[1][l] = bounds_item[1]

                        mu0 = (x_values_fit[-1] + x_values_fit[0])/2
                        sigma0 = (x_values_fit[-1] - x_values_fit[0])/3
                        B0 = (y_values_fit[-1] - y_values_fit[0])/(3*sigma0)
                        C0 = np.min(y_values_fit)
                        A0 = np.max(y_values_fit) - C0
                        p0 = [A0, mu0, sigma0, B0, C0]
                        for i in range(0, len(p0)):
                            if p0[i] < fit_bounds[0][i]:
                                p0[i] = fit_bounds[0][i]
                            elif p0[i] > fit_bounds[1][i]:
                                p0[i] = fit_bounds[1][i]

                        def fitfunc(x, A, mu, sigma, B, C):
                            return A*np.exp(-1/2*((x-mu)**2)/(sigma**2)) + B*x + C
                        popt, pcov = scipy.optimize.curve_fit(fitfunc, x_values_fit, y_values_fit, p0=p0, bounds=fit_bounds)
                        perr = np.sqrt(np.diag(pcov))

                        for popt_i, popt_k in enumerate(fitKeys):
                            popt_map[popt_k] = Basics.roundFloat(popt[popt_i])
                        for perr_i, perr_k in enumerate(fitKeys):
                            perr_map[perr_k] = Basics.roundFloat(perr[perr_i])
                        fit_function = fitfunc
                        area = (popt_map['A'] - popt_map['B']*popt_map['mu'] - popt_map['C'])*np.sqrt(2*np.pi)*popt_map['sigma']
                        popt_map['area'] = Basics.roundFloat(area)
                        fitKeys.append('area')
                    else:
                        print(f'Fit {fit_type.lower()} not implemented')
                        continue


                    if popt is not None:
                        print(f'Fit Parameters for {fit_type}-Fit for {graph_label.String(enumerate=True)}/{fit_label.String(enumerate=True)}:')
                        Basics.dictFormattedPrint([popt_map, perr_map], r'±', keys=fitKeys)
                        print('\n')

                        mask = CreateMask(x_values_plot, fit_plot_range)
                        x_values_fit = x_values_plot[mask]
                        if fit.get("plot-step", None):
                            x_values_fit = np.arange(x_values_fit[0], x_values_fit[-1], fit.get("plot-step", None))
                        style_args = {
                            "color": fit.get("line-color", None),
                            "linewidth": fit.get("line-size", 2),
                            "linestyle": fit.get("line-style", "--"),
                        }
                        ax.plot(x_values_fit, fit_function(x_values_fit, *popt), **style_args)
                        if fit.get("marker", "") != "" and fit.get("marker", "") is not None:
                            style_args = {
                                "linewidth": fit.get("marker-width", 1),
                                "linestyle": "",
                                "marker": fit.get("marker", ""),
                                "markerfacecolor": fit.get("marker-color", None),
                                "markersize": fit.get("marker-size", 1),
                                "markeredgewidth": fit.get("marker-width", 1),
                            }
                            if fit.get("marker-step", None):
                                x_values_fit = np.arange(x_values_fit[0], x_values_fit[-1], fit.get("marker-step", None))

                            ax.plot(x_values_fit, fit_function(x_values_fit, *popt), **style_args)


            if isinstance(graph_data, Plotter.Scatter):
                mask_x = CreateMask(graph_data.x_values, graph_data.x_range)
                mask_y = CreateMask(graph_data.y_values[mask_x], graph_data.y_range)

                markerstyling = graph_config.get("marker", Plotter.Marker())
                ax.scatter(graph_data.x_values[mask_x][mask_y], graph_data.y_values[mask_x][mask_y], zorder=graph_data.z, marker=markerstyling.style, color=markerstyling.color, s=markerstyling.size, alpha=markerstyling.alpha)


        ## Objects
        for i, obj in enumerate(objects):
            if isinstance(obj, Plotter.Objects.Vector):
                ax.quiver(*obj.origin, *(obj.vector*obj.length),
                angles='xy', scale_units='xy',
                scale=1, color=obj.color, alpha=obj.alpha,
                width=obj.width/1000,
                headwidth=obj.headwidth, headlength=obj.headlength
                )
                if obj.mirror:
                    ax.quiver(*obj.origin, *(-obj.vector*obj.length),
                    angles='xy', scale_units='xy',
                    scale=1, color=obj.color, alpha=obj.alpha,
                    width=obj.width/1000,
                    headwidth=obj.headwidth, headlength=obj.headlength
                    )
                Log.Print("Plotted Vector")


        ## Axes
        # Ticks, Tick Labels and Grid
        grid_conf = x_axis_conf.get("major-grid", Plotter.TickGrid().DefaultMajor())
        label_conf = x_axis_conf.get("major-label", Plotter.Label())
        tick_conf = x_axis_conf.get("major-ticker", Plotter.TickTicker().DefaultMajor())
        ax.tick_params(which="major", axis="x", labelcolor = label_conf.color, labelsize = label_conf.fontsize, direction=tick_conf.direction, length=tick_conf.length, width=tick_conf.width, color=tick_conf.color)
        ax.grid(which="major", axis="x", color = grid_conf.color, alpha=grid_conf.alpha, linewidth=grid_conf.linewidth, linestyle=grid_conf.linestyle)

        grid_conf = x_axis_conf.get("minor-grid", Plotter.TickGrid().DefaultMinor())
        label_conf = x_axis_conf.get("minor-label", Plotter.Label())
        tick_conf = x_axis_conf.get("minor-ticker", Plotter.TickTicker().DefaultMinor())
        ax.tick_params(which="minor", axis="x", labelcolor = label_conf.color, labelsize = label_conf.fontsize, direction=tick_conf.direction, length=tick_conf.length, width=tick_conf.width, color=tick_conf.color)
        ax.grid(which="minor", axis="x", color = grid_conf.color, alpha=grid_conf.alpha, linewidth=grid_conf.linewidth, linestyle=grid_conf.linestyle)

        grid_conf = y_axis_conf.get("major-grid", Plotter.TickGrid().DefaultMajor())
        label_conf = y_axis_conf.get("major-label", Plotter.Label())
        tick_conf = y_axis_conf.get("major-ticker", Plotter.TickTicker().DefaultMajor())
        ax.tick_params(which="major", axis="y", labelcolor = label_conf.color, labelsize = label_conf.fontsize, direction=tick_conf.direction, length=tick_conf.length, width=tick_conf.width, color=tick_conf.color)
        ax.grid(which="major", axis="y", color = grid_conf.color, alpha=grid_conf.alpha, linewidth=grid_conf.linewidth, linestyle=grid_conf.linestyle)

        grid_conf = y_axis_conf.get("minor-grid", Plotter.TickGrid().DefaultMinor())
        label_conf = y_axis_conf.get("minor-label", Plotter.Label())
        tick_conf = y_axis_conf.get("minor-ticker", Plotter.TickTicker().DefaultMinor())
        ax.tick_params(which="minor", axis="y", labelcolor = label_conf.color, labelsize = label_conf.fontsize, direction=tick_conf.direction, length=tick_conf.length, width=tick_conf.width, color=tick_conf.color)
        ax.grid(which="minor", axis="y", color = grid_conf.color, alpha=grid_conf.alpha, linewidth=grid_conf.linewidth, linestyle=grid_conf.linestyle)

        
        
        #ax.set_xscale(x_axis_conf.get("scale", "linear"))
        #ax.set_yscale(y_axis_conf.get("scale", "linear"))

        # Limits
        if x_axis_conf.get("limits", None):
            if isinstance(x_axis_conf.get("limits", None), tuple):
                if x_axis_conf.get("limits")[0] > x_axis_conf.get("limits")[1]:
                    print("x-limit start is larger than end. Skipping.")
                else:
                    ax.set_xlim(x_axis_conf.get("limits"))
        if y_axis_conf.get("limits", None):
            if isinstance(y_axis_conf.get("limits", None), tuple):
                if y_axis_conf.get("limits")[0] > y_axis_conf.get("limits")[1]:
                    print("y-limit start is larger than end. Skipping.")
                else:
                    ax.set_ylim(y_axis_conf.get("limits"))
                    
        if y_axis_conf.get("limits", None) in ("x", "x-axis"):
            ax.set_ylim(ax.get_xlim())
        elif x_axis_conf.get("limits", None) in ("y", "y-axis"):
            ax.set_xlim(ax.get_ylim())

        # Ticks
        x_major_ticks = x_axis_conf.get("major-ticks", None)
        if isinstance(x_major_ticks, (tuple, list)):
            ax.xaxis.set_major_locator(FixedLocator(x_major_ticks))
        elif isinstance(x_major_ticks, int):
            min, max = ax.get_xlim()
            ax.xaxis.set_major_locator(FixedLocator(np.linspace(min, max, x_major_ticks, endpoint=True)))
        elif isinstance(x_major_ticks, float):
            ax.xaxis.set_major_locator(MultipleLocator(x_major_ticks))
        else:
            pass
            
        x_minor_ticks = x_axis_conf.get("minor-ticks", None)
        if isinstance(x_minor_ticks, (tuple, list)):
            ax.xaxis.set_minor_locator(FixedLocator(x_minor_ticks))
        elif isinstance(x_minor_ticks, int):
            min, max = ax.get_xlim()
            ax.xaxis.set_minor_locator(FixedLocator(np.linspace(min, max, x_minor_ticks, endpoint=True)))
        elif isinstance(x_minor_ticks, float):
            ax.xaxis.set_minor_locator(MultipleLocator(x_minor_ticks))
        else:
            pass

        y_major_ticks = y_axis_conf.get("major-ticks", None)
        if isinstance(y_major_ticks, (tuple, list)):
            ax.yaxis.set_major_locator(FixedLocator(y_major_ticks))
        elif isinstance(y_major_ticks, int):
            min, max = ax.get_ylim()
            ax.yaxis.set_minor_locator(FixedLocator(np.linspace(min, max, y_major_ticks, endpoint=True)))
        elif isinstance(y_major_ticks, float):
            ax.yaxis.set_major_locator(MultipleLocator(y_major_ticks))
        else:
            pass

        y_minor_ticks = y_axis_conf.get("minor-ticks", None)
        if isinstance(y_minor_ticks, (tuple, list)):
            ax.yaxis.set_minor_locator(FixedLocator(y_minor_ticks))
        elif isinstance(y_minor_ticks, int):
            min, max = ax.get_ylim()
            ax.yaxis.set_minor_locator(FixedLocator(np.linspace(min, max, y_minor_ticks, endpoint=True)))
        elif isinstance(y_minor_ticks, float):
            ax.yaxis.set_minor_locator(MultipleLocator(y_minor_ticks))
        else:
            pass

        if x_axis_conf.get("scale", "linear") == "log":
            ax.set_xscale("log")
        if y_axis_conf.get("scale", "linear") == "log":
            ax.set_yscale("log")

        #Legend
        Legend = plot_conf.get("legend", Plotter.Legend())
        if Legend.show:
            ax.legend(loc=Legend.position)

        count_x_ticks_major = len(ax.get_xticks())
        count_x_ticks_minor = len(ax.get_xticks(minor=True))
        count_y_ticks_major = len(ax.get_yticks())
        count_y_ticks_minor = len(ax.get_yticks(minor=True))
        if count_x_ticks_major > 1000 or count_x_ticks_minor > 1000 or count_y_ticks_major > 1000 or count_y_ticks_minor > 1000 :
            print("Too many Ticks:")
            print(count_x_ticks_major, count_x_ticks_minor, count_y_ticks_major, count_y_ticks_minor)
            return None

        # Plotting
        fig.tight_layout()
        if filename and filename != "":
            dpi = plot_conf.get("dpi", 200)
            fig.savefig(filename, dpi=dpi)
        if Basics.dictMultiGet(plot_conf, ('show', 'print'), True):
            fig.show()
        return fig

    def QuickPlotArray(array: np.ndarray, x: int = 0, y: tuple[int]|list[int]|int = "*", range_x = None, range_y = None, save: str = None) -> plt.Figure:
        dim = array.ndim
        if dim != 2:
            print('Input must be 2D-np.ndarray with atleast 2 columns')
            return None
        
        x_values = Array.Col(array, x)
        y_values = []
        if type(y) is int:
            y_values.append(Array.Col(array, y))
        elif type(y) is tuple:
            for i in y:
                if type(i) != int:
                    print('y-index must be integer for column position')
                    return None
                y_values.append(Array.Col(array, i))
        elif y == "*":
            length = len(array[0])
            for i in range(0, length):
                if i != x:
                    y_values.append(Array.Col(array, i))

        Config = {
            'plot': {
                'title': Plotter.Label('QuickPlot'),
                'legend': Plotter.Legend(False),
                'save': save
            },
            "x-axis": {
                "limits": range_x
            },
            "y-axis": {
                "limits": range_y
            },
        }
        Graphs = []
        for item in y_values:
            conf  = {
                'data': Plotter.Graph().X(x_values).Y(item),
            }
            Graphs.append(conf)

        return Plotter.Plot(Config, Graphs)



class StackedPlotter:
    def __init__(self):
        self.x = None
        self.Graphs = []

        self.Config = {
            'plot': {
                'title': Plotter.Label('QuickPlot'),
                'legend': Plotter.Legend(False),
            },
            "x-axis": {
            },
            "y-axis": {
            },
        }

    def SetConfig(self, Config):
        self.Config = Config
        return self

    def X(self, x: np.ndarray):
        self.x = x
        return self
    
    def StackY(self, y: np.ndarray, color=None, label=None, linestyling=None):
        item = {
            'data': Plotter.Graph().X(self.x).Y(y),
        }
        if linestyling is None:
            linestyling = Plotter.Line()
        if color:
           linestyling.Color(color)


        if label:
            item['label'] =  Plotter.Label(label, latex=False)

        item['line-styling'] = linestyling



        self.Graphs.append(item)
        return self
    
    def GetGraphs(self):
        return self.Graphs

    def Plot(self):
        Plotter.Plot(self.Config, self.Graphs)




def CreateMask(data, limits: tuple):
    if limits is None or limits == (None, None):
        return [True for i in range(0, len(data))]
    elif isinstance(limits[0], (float, int)) and isinstance(limits[1], (float, int)):
        return (data >= limits[0]) & (data <= limits[1])
    elif isinstance(limits[0], (float, int)):
        return (data >= limits[0]) & (data <= np.inf)
    elif isinstance(limits[1], (float, int)):
        return (data >= -np.inf) & (data <= limits[1])
    print(f"INVALID LIMITS FOR MASK: {limits}")
    return [True for i in range(0, len(data))]


class BoolLogger:
    def __init__(self):
        self.log = True

    def Set(self, value):
        self.log = value

    def Print(self, value):
        if self.log:
            print(value)


from typing import Any, Iterable, Union, Optional, Set, List
import numpy as np
import math


class Basics:
    VERSION = 'Kantris.Basics: 0.2.0'
    @staticmethod
    def dictMultiGet(
        data: dict,
        keys: Union[Iterable, tuple, str],
        default: Any = None,
        allow_multikey: bool = True,
        log: bool = True,
        default_None: bool = False
    ) -> Any:
        # Sentinel-Objekt für eindeutige Prüfung
        _MISSING = object()
        if isinstance(keys, str):
            keys = (keys)
        # Liste aller (key, value)-Paare, die im Dict gefunden wurden
        found = [(key, data.get(key, _MISSING)) for key in keys]
        # Filtere nur die existierenden Keys
        valid = [(k, v) for (k, v) in found if v is not _MISSING]
        count = len(valid)

        result = None
        if count == 0:
            result = default

        elif count == 1:
            result = valid[0][1]

        elif count > 1:
            if allow_multikey:
                if log:
                    keys_str = [k for (k, _) in valid]
                    print(f"INFO for Basics.dictMultiGet: Mehrere Keys gefunden: {keys_str}, gebe ersten Wert zurück.")
                result = valid[0][1]
            else:
                if log:
                    keys_str = [k for (k, _) in valid]
                    print(f"ERROR for Basics.dictMultiGet: Mehrere Keys gefunden {keys_str}, 'allow_multikey=False'. Gebe default zurück.")
                result = default
        if result is None and default_None:
            result = default
        return result
    
    @staticmethod
    def dictListFloats(data: dict, n: int = 4):
        String = ''
        for key in data:
            String = String + f'{key}: {Basics.roundFloat(data[key], n)}, '
        return String[:-2]
    
    @staticmethod
    def dictFormattedPrint(dicts, joins: tuple|list = (), keys: tuple|list = None):
        if isinstance(dicts, dict):
            dicts = [dicts, ]
        if isinstance(joins, str):
            joins = [joins, ]
        if len(dicts)-1 != len(joins):
            print(f'ERROR for Basics.dictFormattedPrint: {len(dicts)} dicts given, but only {len(joins)} joins. Need N-1 joins')

        if keys is None:
            keys = set(dicts[0].keys())
        for item in dicts[1:-1]:
            if set(item.keys()) != keys:
                print(f'ERROR for Basics.dictFormattedPrint: All dicts do not have the same Keys')
        
        max_length = max([len(key) for key in keys])
        _MISSING = object()
        for key in keys:
            for i in range(len(dicts)):
                if i == 0:
                    Joined = dicts[i][key]
                else:
                    dict_i_key = dicts[i].get(key, _MISSING)
                    if dict_i_key == _MISSING:
                        continue
                    Joined = f'{Joined} {joins[i-1]} {dict_i_key}'

            print(f'{key}{" " * (max_length - len(key))}: {Joined}')




    @staticmethod
    def roundFloat(x: Union[float, np.floating], sig: int = 4) -> float:
        """
        Rundet x auf 'sig' signifikante Stellen.
        
        Args:
            x: Zahl (float oder np.floating), die gerundet werden soll.
            sig: Anzahl der Signifikanten (default=4).
            
        Returns:
            Gerundeter Wert als Python float.
        """
        if x == np.inf or x == -np.inf:
            return x
        x = float(x)
        if x == 0:
            return 0.0
        # Exponent in Basis 10
        exp = math.floor(math.log10(abs(x)))
        # Anzahl Dezimalstellen, auf die round() angewendet wird
        ndigits = sig - exp - 1
        return round(x, ndigits)
    
    @staticmethod
    def SplitSegments(pts: np.ndarray, max_gap: float) -> List[np.ndarray]:
        """
        Teilt das (M,2)-Array dort auf, wo die Distanz
        zwischen aufeinanderfolgenden Punkten > max_gap ist.
        """
        if pts.shape[0] == 0:
            return []

        # dt = Abstände zwischen aufeinanderfolgenden Punkten
        deltas = np.sqrt(np.sum(np.diff(pts, axis=0)**2, axis=1))
        # Indizes, an denen wir brechen wollen
        breaks = np.nonzero(deltas > max_gap)[0]

        segments = []
        start = 0
        for b in breaks:
            segments.append(pts[start : b+1])
            start = b+1
        # letzte Strecke
        segments.append(pts[start:])
        return segments
    @staticmethod
    def linspaceInteger(min_val, max_val, n):
        """
        Gibt n ganzzahlige Werte mit möglichst gleichem Abstand zwischen min und max zurück.
        
        Parameter:
        ----------
        min_val : int
            Der minimale Wert (inklusive)
        max_val : int
            Der maximale Wert (inklusive)
        n : int
            Anzahl gewünschter Werte

        Rückgabe:
        ---------
        ndarray von int
            n ganzzahlige Werte von min bis max mit möglichst gleichem Abstand
        """
        if n <= 1:
            return np.array([int(round((min_val + max_val) / 2))])
        
        # Erzeuge linspace, dann runde und mache unique
        vals = np.linspace(min_val, max_val, n)
        return np.unique(np.round(vals).astype(int))

class Geometry:
    @staticmethod
    def project_to_plane(data: np.ndarray, basis: np.ndarray) -> np.ndarray:
        """
        Projiziert eine M×N-Datenmenge auf eine 2D-Ebene, definiert durch zwei
        lineare unabhängige N-dimensionale Basisvektoren.

        Args:
            data (np.ndarray): Array der Form (M, N), wobei M die Anzahl der Punkte
                                und N die Dimension ist.
            basis (np.ndarray): Array der Form (N, 2), dessen Spalten die beiden
                                Basisvektoren der Ebene sind.

        Returns:
            np.ndarray: Array der Form (M, 2) mit den Koordinaten der projizierten
                        Punkte in der {basis[:,0], basis[:,1]}-Koordinatenbasis.
        """
        # Prüfen, ob data 2D ist und basis die richtige Form hat
        if data.ndim != 2:
            raise ValueError("`data` muss ein 2D-Array der Form (M, N) sein.")
        M, N = data.shape

        if basis.ndim != 2 or basis.shape != (N, 2):
            raise ValueError("`basis` muss die Form (N, 2) haben, "
                                "mit N = data.shape[1].")

        # Prüfen, dass die beiden Basisvektoren linear unabhängig sind
        if np.linalg.matrix_rank(basis) < 2:
            raise ValueError("Die beiden Basisvektoren müssen linear unabhängig sein.")

        # P ist N×2, Pinv seine Pseudoinverse (2×N)
        P = basis
        Pinv = np.linalg.pinv(P)

        # Daten (M×N) transponieren zu (N×M), dann projizieren zu (2×M)
        coords_2xM = Pinv @ data.T

        # Rückgabe in Form (M×2)
        return coords_2xM.T


from .plotter import Plotter, StackedPlotter
from .PCA import PCA, PCA_Signalum
from .FFT import FFT, CWT
from .Basics import Basics
from .Calc import Calc
from .HTTP import HTTP

from .DataManipulation import DataManipulation
from .DataManipulation import Dataframe, Array


__version__ = "0.2.5"


import numpy as np
import pywt
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm, Normalize
from concurrent.futures import ThreadPoolExecutor

import time
from typing import Tuple, Optional, Any, List

from .DataManipulation import DataManipulation, Array

import logging
logger = logging.getLogger(__name__)



class FFT:
    VERSION = 'Kantris.FFT: 0.1.0'
    def FFT(data: np.ndarray, tol: float = 1e-6, log = False, plot = True, unit: str = 'f', scale: float = 1) -> np.ndarray:
        """
        Berechnet die FFT einer Signalreihe mit (möglicherweise) nicht gleichmäßigen Abszissen.

        Parameters
        ----------
        data : np.ndarray
            Eingabe-Array der Form (N, 2), wobei Spalte 0 die x-Werte (z.B. Zeitpunkte) und
            Spalte 1 die zugehörigen y-Werte (z.B. Signalamplituden) enthält.
        tol : float, optional
            Toleranz für die Prüfung auf Gleichmäßigkeit der x-Abstände (Default: 1e-6).

        Returns
        -------
        np.ndarray
            Array der Form (M, 2) mit Spalte 0 die Frequenzen und Spalte 1 die zugehörigen
            Amplituden (Betragswerte der Fourier-Koeffizienten). M = N//2 + 1 für reelle Signale.
        """
        x = data[:, 0]
        y = data[:, 1]

        # Prüfen auf gleichmäßige Abstände
        dx = np.diff(x)
        if not np.allclose(dx, dx[0], atol=tol):
            logger.info("Input x-values are not evenly spaced: interpolating to uniform grid")
            N = len(x)
            x_uniform = np.linspace(x.min(), x.max(), N)
            # Verwende die angegebene LinearInterpol-Funktion
            y_uniform = DataManipulation.LinearInterpol(x_uniform, x, y)
            sampling_interval = x_uniform[1] - x_uniform[0]
        else:
            y_uniform = y
            sampling_interval = dx[0]

        # FFT auf reellem Signal
        N = len(y_uniform)
        fft_vals = np.fft.rfft(y_uniform)
        freqs = np.fft.rfftfreq(N, d=sampling_interval)
        amplitudes = np.abs(fft_vals)
        if unit == 'T':
            #amplitudes = np.flip(amplitudes[1:-1])
            #freqs = np.flip(1/(freqs[1:-1]))
            freqs = 1/freqs

        if scale != 1:
            freqs = freqs/scale

        print("FFT Result:")
        print(f'{N} Datapoints')
        print(f'Maximum Frequency: {N/(x[-1] - x[0])/2} hz')
        print(f'Minimum T: {1/(N/(x[-1] - x[0])/2)} s')

        # Ergebnis: Spaltenweise Frequenz und Amplitude
        return np.column_stack((freqs, amplitudes))


class CWT:
    VERSION = 'Kantris.CWT: 0.2.5'

    @staticmethod
    def CWT(data,
            wavelet='cmor1-1',
            scales=None,
            tol: float = 1e-6,
            cmap:str = 'viridis',
            figsize: tuple = (3440, 1440),
            dpi: float = 100,
            max_workers: int = 1, 
            yscale = "linear",
            znorm = "linear",
            plot: bool = True,
            period_mode: bool = False,
            limit_by_coi: bool = True,
            save_fig: str = False,
            printlogs: bool = True,
            ):
        """
        Optimierte Continuous Wavelet Transform (CWT) mit Fortschrittsanzeigen:
        - NumPy-Interpolation
        - logarithmische Skalen
        - einmaliges Wavelet-Objekt
        - optional Parallelisierung
        """
        Logger = BoolLogger(printlogs)
        Logger.print("[CWT] Start der CWT")
        # Extrahiere Zeit und Signal
        x = data[:, 0]
        y = data[:, 1]

        # Gleichmäßigkeit prüfen
        dx = np.diff(x)
        if not np.allclose(dx, dx[0], atol=tol):
            Logger.print(f"[CWT] Ungleichmäßige Abstände erkannt. Interpolation auf gleichmäßiges Gitter...")
            N = len(x)
            x_uniform = np.linspace(x.min(), x.max(), N)
            y_uniform = np.interp(x_uniform, x, y)
            dt = x_uniform[1] - x_uniform[0]
            Logger.print(f"[CWT] Interpolation abgeschlossen. Δt={dt:.5f} - {60/dt:.5f}hz, Punkte={N}")
        else:
            x_uniform = x
            y_uniform = y
            dt = dx[0]
            Logger.print(f"[CWT] Gleichmäßige Abstände: Δt={dt:.5f} - {60/dt:.5f}hz, Punkte={len(x)}, T={TtoTime(dt*len(x))}")

        # Skalen festlegen
        if scales is None:
            f_c = pywt.ContinuousWavelet(wavelet).center_frequency
            s_min = 2 * f_c
            s_max = f_c * len(data)
            scales = np.logspace(np.log10(s_min), np.log10(s_max), num=300)
            Logger.print(f"[CWT] Generierte {len(scales)} Skalen")
        else:
            Logger.print(f"[CWT] Nutze {len(scales)} übergebene Skalen")

        # Wavelet-Objekt
        Logger.print(f"[CWT] Wavelet: {wavelet}")

        cwt_start_time = time.time()
        cwtmatr, freqs = pywt.cwt(y, scales, wavelet, sampling_period=dt)
        Logger.print(f"Min: {freqs.min():.6f} Hz  |  Max: {freqs.max():.6f} Hz")
        Logger.print(f"Perioden: {1/freqs.max():.2f} s  bis  {1/freqs.min():.2f} s")
        # --- nach CWT-Berechnung ---
        # --- nach CWT-Berechnung ---
        cwtmatr = np.abs(cwtmatr)
        cwt_end_time = time.time()
        Logger.print(f"[CWT] CWT-Berechnung abgeschlossen in {cwt_end_time-cwt_start_time:.2f}s")
        Logger.print(f"[CWT] Berechnet Frequenzen für {len(freqs)} Bins")

        if not plot:
            return cwtmatr, freqs, {"cwt_timer": cwt_end_time-cwt_start_time}
        # slice inner 100% (oder passe auf 90% falls notwendig)
        n_times = cwtmatr.shape[1]
        t0 = int(np.floor(0.0 * n_times))
        t1 = int(np.ceil(1.0 * n_times))
        t0 = max(0, t0)
        t1 = min(n_times, t1)

        Z = cwtmatr[:, t0:t1]               # shape (n_freqs, t1-t0)
        # prefer x_uniform if you interpolated earlier; otherwise fallback to x
        try:
            x_plot = x_uniform[t0:t1]
        except NameError:
            x_plot = x[t0:t1]
        y_plot = freqs                        # length n_freqs

        # --- COI mask: mask out values outside cone (i.e. near edges) ---
        # center frequency of the wavelet
        cw = pywt.ContinuousWavelet(wavelet)
        center_freq = cw.center_frequency
        # distance to edge (seconds) for each plotted time
        dist_to_edge = np.minimum(x_plot - x_plot[0], x_plot[-1] - x_plot)
        # avoid division by zero
        eps = 1e-12
        # coi frequency threshold per time (Hz): frequencies >= coi_freq are considered safe
        coi_freq = (np.sqrt(2.0) * center_freq) / (dist_to_edge + eps)   # shape (n_times,)
        Logger.print(f"[CWT] COI Min frequency: {min(coi_freq):.2f}hz, Max Period: {1/min(coi_freq):.2f}s")



        # mask: valid if freq >= coi_freq(time)
        # build boolean array shape (n_freqs, n_times)
        # freqs is shape (n_freqs,), coi_freq is (n_times,)
        valid = (y_plot[:, None] >= coi_freq[None, :])

        # Mask out values outside COI (i.e., where valid == False)
        Z_masked = np.ma.array(Z, mask=~valid)

        # For plotting the COI line: only plot where coi_freq is inside the computed freq-range
        fmin, fmax = freqs.min(), freqs.max()
        coi_freq_plot = np.where((coi_freq >= fmin) & (coi_freq <= fmax), coi_freq, np.nan)

        # 1) Z-Scale-Norm mit festen Grenzen (z.B. obere Grenze aus Daten, untere Grenze sinnvoll setzen)
        vals = Z_masked.compressed()   # nur die unmaskierten Werte als 1D-Array
        if vals.size == 0:
            raise ValueError("Alle Werte sind maskiert oder es gibt keine gültigen Werte.")
        
        #Set reasonable Z-limits. Primarly required for the huge amount of near 0 Amplitudes and Strong Ultra-Low-Frequency Amplitudes ruining the log scale
        k1 = 1
        while True:
            vmin = np.percentile(vals, k1)
            #Special Conditions
            if vmin < 1e-2:
                vmin = 1e-2
                break
            #Normal Conditions
            if vmin == 0.0:
                k1 += 1
            else:
                break
        k2 = 1
        while True:
            vmax = np.percentile(vals, 100-k2)
            #Normal Conditions
            if vmax == 0.0:
                k2 += 1
            else:
                break
        
        Logger.print("DEBUG", vmin, k1, vmax, k2)
        if not limit_by_coi:
            Z_masked = Z

        # --- Plot ---
        Logger.print("[CWT] Erstelle Plot...")
        plot_start_time = time.time()

        if figsize[0] > 100 or figsize[1] > 100:
            px = 1/dpi
            figsize = (figsize[0]*px, figsize[1]*px)
        fig, axs = plt.subplots(figsize=figsize, dpi=dpi)
        fig.canvas.draw()


        if znorm.lower() in ("log",):
            znorm = LogNorm(vmin=vmin, vmax=vmax)
        else:
            znorm = Normalize(vmin=vmin, vmax=vmax)
        if period_mode:
            y_plot = 1/y_plot
        # choose shading mode robustly
        if (len(x_plot) == Z_masked.shape[1]) and (len(y_plot) == Z_masked.shape[0]):
            pcm = axs.pcolormesh(x_plot, y_plot, Z_masked, norm=znorm, shading='nearest', cmap=cmap)
        else:
            x_edges = compute_edges(x_plot)
            y_edges = compute_edges(y_plot)
            pcm = axs.pcolormesh(x_edges, y_edges, Z_masked, norm=znorm, shading='auto', cmap=cmap)

        # draw COI line only where inside freq range
        axs.plot(x_plot, coi_freq_plot, color='white', linestyle='--', linewidth=1, label='COI')

        # optionally shade the region outside COI (low frequencies near edges) for clarity:
        # we shade only where coi_freq_plot is finite (i.e. inside f-range).
        # fill between ymin and coi (mask NaNs automatically)
        ymin = fmin if fmin > 0 else (np.min(freqs[freqs > 0]) if np.any(freqs > 0) else 1e-12)
        # use where to avoid drawing fill under NaNs
        finite_mask = ~np.isnan(coi_freq_plot)
        if finite_mask.any():
            axs.fill_between(x_plot[finite_mask], ymin, coi_freq_plot[finite_mask],
                             color='white', alpha=0.4, linewidth=0)

        # finalize axes
        fig.colorbar(pcm, ax=axs)
        axs.set_yscale(yscale)

        Logger.print("freqs:")
        Logger.print(freqs[0])
        Logger.print(freqs[-1])
        

        f_freqs_min, f_freqs_max = min(freqs), max(freqs)

        f_coi_min, f_coi_max = min(coi_freq), max(coi_freq)
        if limit_by_coi:
            f_min, f_max = max(f_coi_min, f_freqs_min), min(f_coi_max, f_freqs_max)
        else:
            f_min, f_max = f_freqs_min, f_freqs_max

        if period_mode:
            axs.set_ylim(1/f_min, 1/f_max)
        else:
            axs.set_ylim(f_min, f_max)

        axs.set_xlabel("Time (s)")
        if period_mode:
            axs.set_ylabel("Period (s)")
        else:
            axs.set_ylabel("Frequency (Hz)")
        axs.set_title(f"Continuous Wavelet Transform (Scaleogram) with {wavelet}")
        axs.legend(loc='upper right', fontsize='small')
        plt.tight_layout()
        plt.show()
        if isinstance(save_fig, str) and save_fig != "":
            fig.savefig(save_fig, dpi=dpi)
        plot_end_time = time.time()
        Logger.print(f"[CWT] Plot abgeschlossen in {plot_end_time - plot_start_time:.2f}s")

        return cwtmatr, freqs, {"cwt_timer": cwt_end_time-cwt_start_time}
    
    @staticmethod
    def DWT(
        data: np.ndarray,
        wavelet: str = 'db4',
        level: Optional[int] = None,
        target_period: Optional[float] = None,   # in seconds, z.B. 24*3600
        tol: float = 1e-6,
        cmap: str = 'viridis',
        figsize: Tuple[float, float] = (10, 6),
        yscale: str = "linear",
        pad: bool = True,
        pad_mode: str = 'reflect',
        normalize: Optional[str] = None,         # default None: show raw first
        to_db: bool = False,
        show_approx: bool = True,                # show approx row by default (useful for long periods)
        return_plot: bool = True,
        max_plot_cols: int = 2000,               # for large N decimate columns for plotting
        clip_percentiles: Optional[Tuple[float,float]] = (1.0,99.0), # optional clipping before normalize
        return_extra: bool = True                # return powers and swt_coeffs for analysis
    ) -> Tuple[np.ndarray, np.ndarray, Optional[Any], Optional[Any], Optional[dict]]:
        """
        Optimized SWT-based DWT helper.
        Returns: dwtmatr (levels x time), freqs (levels), fig, ax, extras(dict)
        extras includes 'powers' (RMS per level) and 'swt_coeffs' if requested.
        """

        # --- input checks ---
        if data.ndim != 2 or data.shape[1] < 2:
            raise ValueError("data must be Nx2: data[:,0]=time, data[:,1]=signal")

        x = data[:, 0].astype(float)
        y = data[:, 1].astype(float)

        # --- uniform grid / dt ---
        dx = np.diff(x)
        if not np.allclose(dx, dx[0], atol=tol):
            Norig = len(x)
            x_uniform = np.linspace(x.min(), x.max(), Norig)
            y_uniform = np.interp(x_uniform, x, y)
            dt = x_uniform[1] - x_uniform[0]
        else:
            x_uniform = x
            y_uniform = y
            dt = dx[0] if len(dx) > 0 else 1.0

        # simple unit check: if dt very small or large, warn user (ms vs s)
        if dt < 1e-3:
            # dt in ms or smaller — warn but continue
            print("[DWT] Warning: dt very small ({}). Are your timestamps in seconds?".format(dt))
        if dt > 1e5:
            print("[DWT] Warning: dt very large ({}). Check time units.".format(dt))

        N = len(y_uniform)
        if N < 2:
            raise ValueError("Signal too short (N < 2).")

        max_possible = int(np.floor(np.log2(N)))
        if max_possible < 1:
            raise ValueError("Signal too short for wavelet decomposition (max level < 1).")

        # --- auto-level from target_period (if provided) ---
        if target_period is not None:
            fs = 1.0 / dt
            f_target = 1.0 / float(target_period)
            est_level = int(round(np.log2(fs / f_target) - 1))
            # clamp to sensible range
            est_level = max(1, min(est_level, max_possible))
            if level is None:
                level = est_level
            else:
                level = min(level, max_possible)
            print(f"[DWT] target_period={target_period}s -> est_level={est_level}, using level={level}")
        else:
            if level is None:
                # pragmatic default: prefer deeper analysis for longer signals but not too deep
                level = min(6, max_possible)
            else:
                level = min(level, max_possible)

        # --- ensure divisibility and pad if needed ---
        block = 2 ** level
        rem = N % block
        if rem != 0:
            if pad:
                pad_len = block - rem
                left = pad_len // 2
                right = pad_len - left
                y_uniform = np.pad(y_uniform, (left, right), mode=pad_mode)
                x_left = x_uniform[0] - np.arange(left, 0, -1) * dt
                x_right = x_uniform[-1] + np.arange(1, right + 1) * dt
                x_uniform = np.concatenate([x_left, x_uniform, x_right])
                N = len(y_uniform)
                print(f"[DWT] Padded by {pad_len} samples ({left}/{right}), new N={N}")
            else:
                # reduce level until divisible
                while level > 0 and (N % (2 ** level) != 0):
                    level -= 1
                if level < 1:
                    raise ValueError("No valid level found. Set pad=True or increase signal length.")
                block = 2 ** level

        # --- compute SWT ---
        try:
            swt_coeffs = pywt.swt(y_uniform, wavelet, level=level)
        except Exception as e:
            raise RuntimeError(f"SWT failed: {e}")

        levels = len(swt_coeffs)
        # allocate float32 matrix for memory savings if large
        dwtmatr = np.zeros((levels, N), dtype=np.float32)
        for i, (cA, cD) in enumerate(swt_coeffs):
            dwtmatr[i, :] = np.abs(cD).astype(np.float32)

        # include approximation optionally
        approx_row = None
        if show_approx:
            approx_row = np.abs(swt_coeffs[-1][0]).astype(np.float32)

        # --- frequency mapping (dyadic approx) ---
        fs = 1.0 / dt if dt != 0 else 1.0
        freqs = np.array([fs / (2 ** (i + 1)) for i in range(levels)], dtype=float)

        # --- energy / RMS per level (useful for debugging) ---
        powers = np.array([np.sqrt(np.mean(dwtmatr[i, :] ** 2)) for i in range(levels)], dtype=float)
        approx_rms = None
        if show_approx:
            approx_rms = float(np.sqrt(np.mean(approx_row ** 2)))

        # --- optional percentil-clipping to avoid single outliers dominating ---
        if clip_percentiles is not None:
            lo_p, hi_p = clip_percentiles
            vmin = np.percentile(dwtmatr, lo_p)
            vmax = np.percentile(dwtmatr, hi_p)
            # clip in-place (but keep dtype)
            dwtmatr = np.clip(dwtmatr, vmin, vmax)

        # --- normalization ---
        eps = 1e-12
        if normalize == 'per_level':
            for i in range(dwtmatr.shape[0]):
                row = dwtmatr[i, :]
                mn, mx = row.min(), row.max()
                if mx - mn > 0:
                    dwtmatr[i, :] = (row - mn) / (mx - mn)
        elif normalize == 'global':
            mn, mx = dwtmatr.min(), dwtmatr.max()
            if mx - mn > 0:
                dwtmatr = (dwtmatr - mn) / (mx - mn)

        if to_db:
            dwtmatr = 20.0 * np.log10(dwtmatr + eps)

        # --- prepare for plotting: low freq bottom ---
        dwtmatr_plot = dwtmatr[::-1, :]
        freqs_plot = freqs[::-1]

        if show_approx and approx_row is not None:
            dwtmatr_plot = np.vstack([dwtmatr_plot, approx_row])
            freqs_plot = np.concatenate([freqs_plot, [fs / (2 ** levels)]])

        # --- decimate columns for plotting if large ---
        x_rel = x_uniform - x_uniform[0]   # relative time (s) -> nicer axis
        if return_plot:
            plot_N = dwtmatr_plot.shape[1]
            if plot_N > max_plot_cols:
                # choose indices uniformly
                idx = np.linspace(0, plot_N - 1, max_plot_cols).astype(int)
                x_plot = x_rel[idx]
                Z_plot = dwtmatr_plot[:, idx]
            else:
                x_plot = x_rel
                Z_plot = dwtmatr_plot

            fig, ax = plt.subplots(figsize=figsize)
            # imshow is usually faster; use extent so x axis corresponds to time and y to freqs
            extent = [x_plot[0], x_plot[-1], freqs_plot[0], freqs_plot[-1]]
            im = ax.imshow(
                Z_plot,
                aspect='auto',
                origin='lower',
                extent=extent,
                cmap=cmap,
                interpolation='nearest'
            )
            cbar = fig.colorbar(im, ax=ax)
            cbar.set_label("Amplitude (abs or normalized)" if not to_db else "Amplitude (dB)")
            if yscale == "log":
                ax.set_yscale("log")   # might raise for imshow; you can set yticks manually instead
            ax.set_xlabel("Time (s) (relative)")
            ax.set_ylabel("Frequency (Hz)")
            ax.set_title(f"Discrete Wavelet (SWT) — wavelet={wavelet}, levels={levels}")
            plt.tight_layout()
            plt.show()
        else:
            fig, ax = None, None

        extras = None
        if return_extra:
            extras = {
                "powers": powers,
                "approx_rms": approx_rms,
                "swt_coeffs": swt_coeffs
            }

        return dwtmatr, freqs, fig, ax, extras
    
    @staticmethod
    def splitCWT(
        data: np.ndarray,
        frequency_range: tuple = "auto",
        n_scales: int = 400,
        wavelet: str = "cmor1-1",
        y_scale: str = "log",
        z_scale: str= "log",
        do_plot: bool = True,
        show_plot: bool = True,
        split_plots: bool = False,
        print_logs: bool = True,
        tol: float = 1e-6,
        plot_config: dict = {},
    ) -> dict:
        """
        Performs a Multi-Range CWT. If the minimal Cone if Influence (COI) frequency of the data (depends on dt) is lower than the critcal frequency ()
        """
        #

        Logger = BoolLogger(print_logs)
        Logger.print("[splitCWT] Start der splitCWT")
        plot = None

        timer = Timer()
        EndResult = {
            'data': {},
            'timers': {},
            'plot': None,
        }
        #
        Timestamps = Array.Col(data, 0)
        Values = Array.Col(data, 1)
        #

        timer.Start("interpolation")
        # Gleichmäßigkeit prüfen
        dx = np.diff(Timestamps)
        if not np.allclose(dx, dx[0], atol=tol):
            Logger.print(f"[splitCWT] Ungleichmäßige Abstände erkannt. Interpolation auf gleichmäßiges Gitter...")
            N = len(Timestamps)
            x_uniform = np.linspace(Timestamps.min(), Timestamps.max(), N)
            y_uniform = np.interp(x_uniform, Timestamps, Values)
            dt = x_uniform[1] - x_uniform[0]
            t_operation = timer.Stop("interpolation")
            Logger.print(f"[splitCWT] Interpolation abgeschlossen. Δt={dt:.5g} - {60/dt:.5g}hz, Punkte={N}, T={TtoTime(dt*len(x_uniform))} in {TtoTime(t_operation)}")
        else:
            x_uniform = Timestamps
            y_uniform = Values
            dt = dx[0]
            t_operation = timer.Stop("interpolation")
            Logger.print(f"[splitCWT] Gleichmäßige Abstände: Δt={dt:.5g} - {60/dt:.5g}hz, Punkte={len(x_uniform)}, T={TtoTime(dt*len(x_uniform))} in {TtoTime(t_operation)}")
        
        Timestamps = x_uniform
        Values = y_uniform
        N_data = len(Timestamps)
        T_in_seconds = Timestamps[-1] - Timestamps[0]
        T_in_days = T_in_seconds/(3600*24)
        # Possible Frequency Ranges
        f_max = 1/dt/2
        f_min_coi = (2*np.sqrt(2*CWT.getBfromWavelet(wavelet))*pywt.ContinuousWavelet(wavelet).center_frequency)/(T_in_seconds)
        f_min = f_min_coi
        if frequency_range == "auto":
            pass
        else:
            if f_max > max(frequency_range):
                f_max =  max(frequency_range)
            else:
                print(f"[splitCWT] Maximum given Frequency of {max(frequency_range):.4g}hz is higher than maximum possible Frequency for data {f_max:.4g}hz (Nisq Limit). Limiting Range.")
            if f_min_coi < min(frequency_range):
                f_min =  min(frequency_range)
            else:
                print(f"[splitCWT] Minium given Frequency of {min(frequency_range):.4g}hz is lower than minimum possible Frequency for data {f_min:.4g}hz (COI Limit). Limiting Range.")


        Logger.print(f'[splitCWT] Maximum Frequency: {f_max:.4g}hz, Minimum Period: {TtoTime(1/f_max)}')
        Logger.print(f'[splitCWT] Minimum Frequency: {f_min:.4g}hz, Maximum Period: {TtoTime(1/f_min)}')

        if dt/T_in_days < 10:
            hasSplit = True
            dt_2 = T_in_days*20
            Logger.print(f'[splitCWT] Δt of {dt:.5g}s is too low for minimum COI Frequency ({TtoTime(T_in_seconds)} would require Δt of {T_in_days*10:.4g}s). Splitting into Ranges [{f_max:.4g}hz to {1/dt_2/2:.4g}hz, Δt={TtoTime(dt)}] and [{1/dt_2/2:.4g}hz to {f_min:.4g}hz, Δt={TtoTime(dt_2)}] to prevent long calculation time.')
            Timestamps_2 = np.linspace(Timestamps[0], Timestamps[-1], int(T_in_seconds/dt_2))
            Values_2 = np.interp(Timestamps_2, Timestamps, Values)
        else:
            hasSplit = False

        if hasSplit:
            #Notes: freqs1[0] ~ freqs2[-1]
            #
            #
            ## CWT1


            timer.Start("CWT1")
            Logger.print(f'[splitCWT] starting CWT-1 calculation with {wavelet}-Wavelet and {int(n_scales)}-Scales')
            
            freqs_1 = np.logspace(np.log10(1/dt_2/2), np.log10(f_max), int(n_scales))
            central = pywt.central_frequency(wavelet)
            scales = central / (freqs_1 * dt)
            cwtmatr_1, freqs_1 = pywt.cwt(Values, scales, wavelet, sampling_period=dt)
            cwtmatr_abs_1 = np.abs(cwtmatr_1)

            t_operation = timer.Stop("CWT1")
            Logger.print(f"[splitCWT] CWT-1 calculation done in {TtoTime(t_operation)}")

            ## CWT2
            timer.Start("CWT2")
            Logger.print(f'[splitCWT] starting CWT-2 calculation with {wavelet}-Wavelet and {int(n_scales)}-Scales')
            
            freqs_2 = np.logspace(np.log10(f_min), np.log10(1/dt_2/2), n_scales)
            central = pywt.central_frequency(wavelet)
            scales = central / (freqs_2 * dt_2)
            cwtmatr_2, freqs_2 = pywt.cwt(Values_2, scales, wavelet, sampling_period=dt_2)
            cwtmatr_abs_2 = np.abs(cwtmatr_2)

            t_operation = timer.Stop("CWT2")
            Logger.print(f"[splitCWT] CWT-2 calculation done in {TtoTime(t_operation)}")

            if do_plot:
                if not split_plots:
                    timer.Start("plot")
                    plot = CWT.PlotMultiCWT([cwtmatr_abs_1, cwtmatr_abs_2], [freqs_1, freqs_2], [Timestamps, Timestamps_2], y_scale=y_scale, z_scale=z_scale, wavelet=wavelet, inter=plot_config.get('inter', 5))
                    t_operation = timer.Stop("plot")
                    Logger.print(f"[splitCWT] Generating Plot done in {TtoTime(t_operation)}")
                    plot.show()
                else:
                    timer.Start("plot")
                    plot_1 = CWT.PlotCWT(cwtmatr_abs_1, freqs_1, Timestamps, y_scale=y_scale, z_scale=z_scale, wavelet=wavelet, inter=plot_config.get('inter', 5))
                    plot_1.show()
                    plot_2 = CWT.PlotCWT(cwtmatr_abs_2, freqs_2, Timestamps_2, y_scale=y_scale, z_scale=z_scale, wavelet=wavelet, inter=plot_config.get('inter', 5))
                    plot_2.show()
                    t_operation = timer.Stop("plot")
                    Logger.print(f"[splitCWT] Generating Plot done in {TtoTime(t_operation)}")

        elif not hasSplit:
            timer.Start("CWT")
            Logger.print(f'[splitCWT] starting CWT calculation with {wavelet}-Wavelet and {n_scales}-Scales')
            freqs = np.logspace(np.log10(f_min), np.log10(f_max), n_scales)
            central = pywt.central_frequency(wavelet)
            scales = central / (freqs * dt)
            cwtmatr, freqs = pywt.cwt(Values, scales, wavelet, sampling_period=dt)
            cwtmatr_abs = np.abs(cwtmatr)
            t_operation = timer.Stop("CWT")
            Logger.print(f"[splitCWT] CWT-calculation done in {TtoTime(t_operation)}")
            if do_plot:
                timer.Start("plot")
                plot = CWT.PlotCWT(cwtmatr_abs, freqs, Timestamps, y_scale=y_scale, z_scale=z_scale, wavelet=wavelet, inter=plot_config.get('inter', 5))
                t_operation = timer.Stop("plot")
                Logger.print(f"[splitCWT] Generating Plot done in {TtoTime(t_operation)}")
                plot.show()
        
        EndResult['plot'] = plot
        return EndResult



    @staticmethod
    def PlotCWT(
        cwtmatr: np.ndarray, freqs: np.ndarray, timestamps: np.ndarray, limit_by_coi: bool=True, wavelet: str = None, y_scale="log", z_scale="linear",
        cmap="viridis", figsize=(12,6), dpi=400, inter=5, cwtmatr_absolute: bool = True) -> plt.Figure:
        if not cwtmatr_absolute:
            cwtmatr = np.abs(cwtmatr)
        Z_values = cwtmatr[:, 0:cwtmatr.shape[1]]

        if limit_by_coi and wavelet is None:
            print("NO WAVELET GIVEN")
            raise Exception
            return

        if limit_by_coi:
            x_plot = timestamps
            y_plot = freqs
            # distance to edge (seconds) for each plotted time
            dist_to_edge = np.minimum(x_plot - x_plot[0], x_plot[-1] - x_plot)
            # avoid division by zero
            eps = 1e-12
            # coi frequency threshold per time (Hz): frequencies >= coi_freq are considered safe
            coi_freq = (np.sqrt(2.0*CWT.getBfromWavelet(wavelet)) * pywt.ContinuousWavelet(wavelet).center_frequency) / (dist_to_edge + eps)   # shape (n_times,)
            # mask: valid if freq >= coi_freq(time)
            # build boolean array shape (n_freqs, n_times)
            # freqs is shape (n_freqs,), coi_freq is (n_times,)
            valid = (y_plot[:, None] >= coi_freq[None, :])
            # Mask out values outside COI (i.e., where valid == False)
            Z_masked = np.ma.array(Z_values, mask=~valid)
            # For plotting the COI line: only plot where coi_freq is inside the computed freq-range
            fmin, fmax = freqs.min(), freqs.max()
            coi_freq_plot = np.where((coi_freq >= fmin) & (coi_freq <= fmax), coi_freq, np.nan)

            vmin = np.percentile(Z_values[valid], 2)
            if vmin < 1e-3:
                vmin = 1e-3
            vmax = np.percentile(Z_values[valid], 98)
        else:
            vmin = np.percentile(Z_values, 2)
            if vmin < 1e-3:
                vmin = 1e-3
            vmax = np.percentile(Z_values, 98)

        if z_scale.lower() in ("log",):
            znorm = LogNorm(vmin=vmin, vmax=vmax)
        else:
            znorm = Normalize(vmin=vmin, vmax=vmax)


        #Create Plot Instance
        fig, axs = plt.subplots(figsize=figsize, dpi=dpi)
        fig.canvas.draw()
        if limit_by_coi:
            if (len(timestamps) == Z_masked.shape[1]) and (len(y_plot) == Z_masked.shape[0]):
                x,y,z = CWT.downsample_x(x_plot, y_plot, Z_masked, inter=inter)
                pcm = axs.pcolormesh(x,y,z, norm=znorm, shading='nearest', cmap=cmap)
            else:
                x_edges = compute_edges(x_plot)
                y_edges = compute_edges(y_plot)
                x,y,z = CWT.downsample_x(x_edges, y_edges, Z_masked, inter=inter)
                pcm = axs.pcolormesh(x,y,z, norm=znorm, shading='auto', cmap=cmap)
            # draw COI line only where inside freq range
            axs.plot(x_plot, coi_freq_plot, color='white', linestyle='--', linewidth=1, label='COI')
            # optionally shade the region outside COI (low frequencies near edges) for clarity:
            # we shade only where coi_freq_plot is finite (i.e. inside f-range).
            # fill between ymin and coi (mask NaNs automatically)
            ymin = fmin if fmin > 0 else (np.min(freqs[freqs > 0]) if np.any(freqs > 0) else 1e-12)
            # use where to avoid drawing fill under NaNs
            finite_mask = ~np.isnan(coi_freq_plot)
            if finite_mask.any():
                axs.fill_between(x_plot[finite_mask], ymin, coi_freq_plot[finite_mask],
                                color='white', alpha=0.4, linewidth=0)
        else:
            pcm = axs.pcolormesh(timestamps, freqs, Z_values, norm=znorm, shading='nearest', cmap=cmap)

        if y_scale in ("linear", ):
            axs.set_yscale("linear")
        else:
            axs.set_yscale("log")
        axs.set_xlabel("Time (s)")
        axs.set_ylabel("Frequency (Hz)")
        axs.set_title(f"Continuous Wavelet Transform (Scaleogram) | {wavelet} | x{inter}")
        fig.colorbar(pcm, ax=axs)
        fig.tight_layout()
        return fig

    @staticmethod
    def PlotMultiCWT(
        cwtmatr: List[np.ndarray], freqs: List[np.ndarray], timestamps: List[np.ndarray], limit_by_coi: bool=True, wavelet: str = None, y_scale="log", z_scale="linear",
        cmap="viridis", figsize=(12,6), dpi=200, cwtmatr_absolute: bool = True, inter: int = 5,
        ) -> plt.Figure:
        if (len(cwtmatr) != len(freqs)) or (len(freqs) != len(timestamps)):
            raise Exception("Inputs for CwtMatrix, Freqs and Timestmaps are not of equal length")
        else:
            PlotDim= len(cwtmatr)
            if PlotDim > 2:
                print(f"[splitCWT] PlotDim is {PlotDim}>2. Limited to 2.")
        if not cwtmatr_absolute:
            cwtmatr = [np.abs(matrix) for matrix in cwtmatr]
        for i in range(0, PlotDim):
            freqs[i], cwtmatr[i] = CWT.__ensure_sorted(freqs[i], cwtmatr[i])

        dt1 = np.median(np.diff(timestamps[0]))
        dt2 = np.median(np.diff(timestamps[1]))
        dt_min = min(dt1, dt2)
        t_start = max(min(timestamps[0]), min(timestamps[1]))  # oder min() wenn beide gleichbereich
        t_end   = min(max(timestamps[0]), max(timestamps[1]))  # sicherstellen Überlappung

        time_common = np.arange(t_start, t_end + dt_min/2, dt_min)
        def interp_to_common(time_orig, A):
            # A shape (nfreq, ntime_orig)
            return np.array([np.interp(time_common, time_orig, row) for row in A])
        
        Ai = [None, None]
        Ai[0] = interp_to_common(timestamps[0], cwtmatr[0])
        Ai[1] = interp_to_common(timestamps[1], cwtmatr[1])
        freqs_all = np.concatenate([freqs[0], freqs[1]])
        A_all     = np.vstack([Ai[0], Ai[1]])
        
        order = np.argsort(freqs_all)
        freqs_sorted = freqs_all[order]
        A_sorted     = A_all[order, :]

        p99 = np.percentile(A_sorted, 99, axis=1, keepdims=True)
        p99[p99 == 0] = 1e-12
        A_norm = A_sorted / p99

        return CWT.PlotCWT(A_norm, freqs_sorted, time_common, limit_by_coi=limit_by_coi, wavelet=wavelet, y_scale=y_scale, z_scale=z_scale, cmap=cmap, figsize=figsize, dpi=dpi, cwtmatr_absolute=cwtmatr_absolute, inter=inter)

    @staticmethod
    def __ensure_sorted(freqs, mat):
        # sort nach freq aufsteigend
        idx = np.argsort(freqs)
        freqs_s = freqs[idx]
        mat_s = mat[idx, :]
        return freqs_s, mat_s
    
    @staticmethod
    def getBfromWavelet(wavelet: str):
        if wavelet[0:4] == "cmor":
            k = wavelet.strip("cmor").split('-')[0]
            return float(k)
        else:
            return 1.0

    @staticmethod
    def downsample_x(x, y, Z, inter=5):
        """
        Downsample Z only along the x-axis by grouping 'inter' columns together.
        Returns (x_ds, y, Z_ds).
        
        Parameters
        ----------
        x : 1D array-like, length nx (columns of Z)
        y : 1D array-like, length ny (rows of Z)
        Z : 2D array-like or numpy.ma.MaskedArray with shape (ny, nx)
        inter : int, downsampling factor (e.g. 10 => each group of 10 x points -> 1)
        
        Returns
        -------
        x_ds : 1D numpy array length m = ceil(nx / inter)
        y     : same as input y (unchanged)
        Z_ds  : 2D array (ny, m) or MaskedArray if input was masked
        """
        x = np.asarray(x)
        y = np.asarray(y)

        # Accept masked arrays, preserve mask behavior
        is_masked = np.ma.is_masked(Z) or isinstance(Z, np.ma.MaskedArray)
        if is_masked:
            Z = np.ma.asarray(Z)
        else:
            Z = np.asarray(Z)

        if Z.ndim != 2:
            raise ValueError("Z must be 2D (ny, nx).")
        ny, nx = Z.shape
        if x.shape[0] != nx:
            raise ValueError(f"len(x) ({x.shape[0]}) != Z.shape[1] ({nx})")
        if y.shape[0] != ny:
            raise ValueError(f"len(y) ({y.shape[0]}) != Z.shape[0] ({ny})")

        m = (nx + inter - 1) // inter  # number of output x bins
        # compute downsampled x as block mean
        x_ds = np.empty(m, dtype=x.dtype)
        if is_masked:
            col_means = []
            for j in range(m):
                start = j * inter
                end = min(start + inter, nx)
                x_ds[j] = x[start:end].mean()
                # mean across the sliced columns for each row (preserves mask semantics)
                col_mean = Z[:, start:end].mean(axis=1)
                col_means.append(col_mean)
            Z_ds = np.ma.column_stack(col_means)   # shape (ny, m)
        else:
            # vectorized building via list comprehension (fast enough)
            cols = []
            for j in range(m):
                start = j * inter
                end = min(start + inter, nx)
                x_ds[j] = x[start:end].mean()
                cols.append(Z[:, start:end].mean(axis=1))
            Z_ds = np.column_stack(cols)  # shape (ny, m)

        return x_ds, y, Z_ds

def compute_edges(a: np.ndarray) -> np.ndarray:
    """Compute edge array from centers (length -> length+1)."""
    a = np.asarray(a, dtype=float)
    if a.size < 2:
        return np.array([a[0] - 0.5, a[0] + 0.5])
    mid = 0.5 * (a[:-1] + a[1:])
    first = a[0] - 0.5 * (a[1] - a[0])
    last  = a[-1] + 0.5 * (a[-1] - a[-2])
    edges = np.empty(a.size + 1, dtype=float)
    edges[0] = first
    edges[-1] = last
    edges[1:-1] = mid
    return edges

class BoolLogger():
    def __init__(self, do: bool=True):
        self.do = do
    def print(self, *args):
        if self.do:
            print(*args)
        else:
            pass

class Timer():
    def __init__(self):
        self.timers = {}

    def Start(self, name: str):
        self.timers[name] = [time.time(), None]

    def Stop(self, name: str):
        times = self.timers.get(name, None)
        if times is None:
            return None
        else:
            self.timers[name][1] = time.time()
        return self.timers[name][1] - self.timers[name][0]
    
    def StopAndPrint(self, name: str):
        times = self.timers.get(name, None)
        if times is None:
            return None
        else:
            self.timers[name][1] = time.time()
        print(TtoTime(self.timers[name][1] - self.timers[name][0]))
        return self.timers[name][1] - self.timers[name][0]

    def Get(self, name: str):
        return self.timers.get(name, None)
    def GetAll(self):
        return self.timers

def TtoTime(T) -> str:
    t = np.abs(T)
    if t < 1:
        return f'{t*1000:.4g}ms'
    elif t < 60*2:
        return f'{t:.4g}s'
    elif t < 3600*2:
        return f'{t/60:.4g}m'
    elif t < 3600*24*2:
        return f'{t/3600:.4g}h'
    elif t < 3600*24*365*2:
        return f'{t/3600/24:.4g}d'
    else:
        return f'{t/3600/24/365:.4g}a'
    

